@@ -2,7 +2,7 @@
  * paths.c:   a path manipulation library using svn_stringbuf_t
  *
  * ====================================================================
- * Copyright (c) 2000-2004 CollabNet.  All rights reserved.
+ * Copyright (c) 2000-2007, 2009 CollabNet.  All rights reserved.
  *
  * This software is licensed as described in the file COPYING, which
  * you should have received as part of this distribution.  The terms
@@ -23,14 +23,17 @@
 
 #include <apr_file_info.h>
 #include <apr_lib.h>
+#include <apr_uri.h>
 
 #include "svn_string.h"
+#include "svn_dirent_uri.h"
 #include "svn_path.h"
 #include "svn_private_config.h"         /* for SVN_PATH_LOCAL_SEPARATOR */
 #include "svn_utf.h"
 #include "svn_io.h"                     /* for svn_io_stat() */
 #include "svn_ctype.h"
 
+#include "private_uri.h"
 
 /* The canonical empty path.  Can this be changed?  Well, change the empty
    test below and the path library will work, not so sure about the fs/wc
@@ -49,56 +52,53 @@
 const char *
 svn_path_internal_style(const char *path, apr_pool_t *pool)
 {
-  if ('/' != SVN_PATH_LOCAL_SEPARATOR)
-    {
-      char *p = apr_pstrdup(pool, path);
-      path = p;
-
-      /* Convert all local-style separators to the canonical ones. */
-      for (; *p != '\0'; ++p)
-        if (*p == SVN_PATH_LOCAL_SEPARATOR)
-          *p = '/';
-    }
-
-  return svn_path_canonicalize(path, pool);
-  /* FIXME: Should also remove trailing /.'s, if the style says so. */
+#if defined(WIN32) || defined(__CYGWIN__)
+  if ((path[0] == '/' || path[0] == '\\') && path[1] == path[0])
+    return svn_dirent_internal_style(path, pool);
+#endif
+  return svn_uri_internal_style(path, pool);
 }
 
 
 const char *
 svn_path_local_style(const char *path, apr_pool_t *pool)
 {
-  path = svn_path_canonicalize(path, pool);
-  /* FIXME: Should also remove trailing /.'s, if the style says so. */
-
-  /* Internally, Subversion represents the current directory with the
-     empty string.  But users like to see "." . */
-  if (SVN_PATH_IS_EMPTY(path))
-    return ".";
-
-  if ('/' != SVN_PATH_LOCAL_SEPARATOR)
-    {
-      char *p = apr_pstrdup(pool, path);
-      path = p;
-
-      /* Convert all canonical separators to the local-style ones. */
-      for (; *p != '\0'; ++p)
-        if (*p == '/')
-          *p = SVN_PATH_LOCAL_SEPARATOR;
-    }
-
-  return path;
+#if defined(WIN32) || defined(__CYGWIN__)
+  if (path[0] == '/' && path[1] == '/')
+    return svn_dirent_local_style(path, pool);
+#endif
+  return svn_uri_local_style(path, pool);
 }
 
-
 
 #ifndef NDEBUG
+/* This function is an approximation of svn_path_is_canonical.
+ * It is supposed to be used in functions that do not have access
+ * to a pool, but still want to assert that a path is canonical.
+ *
+ * PATH with length LEN is assumed to be canonical if it isn't
+ * the platform's empty path (see definition of SVN_PATH_IS_PLATFORM_EMPTY),
+ * and does not contain "/./", and any one of the following
+ * conditions is also met:
+ *
+ *  1. PATH has zero length
+ *  2. PATH is the root directory (what exactly a root directory is
+ *                                depends on the platform)
+ *  3. PATH is not a root directory and does not end with '/'
+ *
+ * If possible, please use svn_path_is_canonical instead.
+ */
 static svn_boolean_t
 is_canonical(const char *path,
              apr_size_t len)
 {
   return (! SVN_PATH_IS_PLATFORM_EMPTY(path, len)
-          && (len <= 1 || path[len-1] != '/'));
+          && strstr(path, "/./") == NULL
+          && (len == 0
+              || svn_dirent_is_root(path, len)
+              /* The len > 0 check is redundant, but here to make
+               * sure we never ever end up indexing with -1. */
+              || (len > 0 && path[len-1] != '/')));
 }
 #endif
 
@@ -111,8 +111,8 @@ char *svn_path_join(const char *base,
   apr_size_t clen = strlen(component);
   char *path;
 
-  assert(is_canonical(base, blen));
-  assert(is_canonical(component, clen));
+  assert(svn_path_is_canonical(base, pool));
+  assert(svn_path_is_canonical(component, pool));
 
   /* If the component is absolute, then return it.  */
   if (*component == '/')
@@ -152,7 +152,7 @@ char *svn_path_join_many(apr_pool_t *pool, const char *base, ...)
 
   total_len = strlen(base);
 
-  assert(is_canonical(base, total_len));
+  assert(svn_path_is_canonical(base, pool));
 
   if (total_len == 1 && *base == '/')
     base_is_root = TRUE;
@@ -172,7 +172,7 @@ char *svn_path_join_many(apr_pool_t *pool, const char *base, ...)
     {
       len = strlen(s);
 
-      assert(is_canonical(s, len));
+      assert(svn_path_is_canonical(s, pool));
 
       if (SVN_PATH_IS_EMPTY(s))
         continue;
@@ -281,7 +281,7 @@ svn_path_component_count(const char *path)
         ++path;
 
       start = path;
-      
+
       while (*path && *path != '/')
         ++path;
 
@@ -292,6 +292,7 @@ svn_path_component_count(const char *path)
   return count;
 }
 
+
 /* Return the length of substring necessary to encompass the entire
  * previous path segment in PATH, which should be a LEN byte string.
  *
@@ -315,14 +316,15 @@ previous_segment(const char *path,
     return len;
 }
 
+
 void
-svn_path_add_component(svn_stringbuf_t *path, 
+svn_path_add_component(svn_stringbuf_t *path,
                        const char *component)
 {
   apr_size_t len = strlen(component);
 
   assert(is_canonical(path->data, path->len));
-  assert(is_canonical(component, len));
+  assert(is_canonical(component, strlen(component)));
 
   /* Append a dir separator, but only if this path is neither empty
      nor consists of a single dir separator already. */
@@ -346,6 +348,7 @@ svn_path_remove_component(svn_stringbuf_t *path)
   path->data[path->len] = '\0';
 }
 
+
 void
 svn_path_remove_components(svn_stringbuf_t *path, apr_size_t n)
 {
@@ -362,7 +365,7 @@ svn_path_dirname(const char *path, apr_pool_t *pool)
 {
   apr_size_t len = strlen(path);
 
-  assert(is_canonical(path, len));
+  assert(svn_path_is_canonical(path, pool));
 
   return apr_pstrmemdup(pool, path, previous_segment(path, len));
 }
@@ -374,7 +377,7 @@ svn_path_basename(const char *path, apr_pool_t *pool)
   apr_size_t len = strlen(path);
   apr_size_t start;
 
-  assert(is_canonical(path, len));
+  assert(svn_path_is_canonical(path, pool));
 
   if (len == 1 && path[0] == '/')
     start = 0;
@@ -389,7 +392,6 @@ svn_path_basename(const char *path, apr_pool_t *pool)
 }
 
 
-
 void
 svn_path_split(const char *path,
                const char **dirpath,
@@ -409,7 +411,7 @@ svn_path_split(const char *path,
 int
 svn_path_is_empty(const char *path)
 {
-  /* assert (is_canonical (path, strlen (path))); ### Expensive strlen */
+  assert(is_canonical(path, strlen(path)));
 
   if (SVN_PATH_IS_EMPTY(path))
     return 1;
@@ -417,7 +419,6 @@ svn_path_is_empty(const char *path)
   return 0;
 }
 
-
 int
 svn_path_compare_paths(const char *path1,
                        const char *path2)
@@ -427,8 +428,8 @@ svn_path_compare_paths(const char *path1,
   apr_size_t min_len = ((path1_len < path2_len) ? path1_len : path2_len);
   apr_size_t i = 0;
 
-  assert(is_canonical(path1, path1_len));
-  assert(is_canonical(path2, path2_len));
+  assert(is_canonical(path1, strlen(path1)));
+  assert(is_canonical(path2, strlen(path2)));
 
   /* Skip past common prefix. */
   while (i < min_len && path1[i] == path2[i])
@@ -436,7 +437,7 @@ svn_path_compare_paths(const char *path1,
 
   /* Are the paths exactly the same? */
   if ((path1_len == path2_len) && (i >= min_len))
-    return 0;    
+    return 0;
 
   /* Children of paths are greater than their parents, but less than
      greater siblings of their parents. */
@@ -456,191 +457,40 @@ svn_path_compare_paths(const char *path1,
   return (unsigned char)(path1[i]) < (unsigned char)(path2[i]) ? -1 : 1;
 }
 
-
-/* Return the string length of the longest common ancestor of PATH1 and PATH2.  
- *
- * This function handles everything except the URL-handling logic 
- * of svn_path_get_longest_ancestor, and assumes that PATH1 and 
- * PATH2 are *not* URLs.  
- *
- * If the two paths do not share a common ancestor, return 0. 
- *
- * New strings are allocated in POOL.
- */
-static apr_size_t
-get_path_ancestor_length(const char *path1,
-                         const char *path2,
-                         apr_pool_t *pool)
-{
-  apr_size_t path1_len, path2_len;
-  apr_size_t i = 0;
-  apr_size_t last_dirsep = 0;
-  
-  path1_len = strlen(path1);
-  path2_len = strlen(path2);
-
-  if (SVN_PATH_IS_EMPTY(path1) || SVN_PATH_IS_EMPTY(path2))
-    return 0;
-
-  while (path1[i] == path2[i])
-    {
-      /* Keep track of the last directory separator we hit. */
-      if (path1[i] == '/')
-        last_dirsep = i;
-
-      i++;
-
-      /* If we get to the end of either path, break out. */
-      if ((i == path1_len) || (i == path2_len))
-        break;
-    }
-
-  /* last_dirsep is now the offset of the last directory separator we
-     crossed before reaching a non-matching byte.  i is the offset of
-     that non-matching byte. */
-  if (((i == path1_len) && (path2[i] == '/'))
-           || ((i == path2_len) && (path1[i] == '/'))
-           || ((i == path1_len) && (i == path2_len)))
-    return i;
-  else
-    return last_dirsep;
-}
-
-
 char *
 svn_path_get_longest_ancestor(const char *path1,
                               const char *path2,
                               apr_pool_t *pool)
 {
-  svn_boolean_t path1_is_url, path2_is_url;
-  path1_is_url = svn_path_is_url(path1);
-  path2_is_url = svn_path_is_url(path2);
-
-  if (path1_is_url && path2_is_url) 
-    {
-      apr_size_t path_ancestor_len; 
-      apr_size_t i = 0;
-
-      /* Find ':' */
-      while (1)
-        {
-          /* No shared protocol => no common prefix */
-          if (path1[i] != path2[i])
-            return apr_pmemdup(pool, SVN_EMPTY_PATH, 
-                               sizeof(SVN_EMPTY_PATH));
-
-          if (path1[i] == ':') 
-            break;
-
-          /* They're both URLs, so EOS can't come before ':' */
-          assert((path1[i] != '\0') && (path2[i] != '\0'));
-
-          i++;
-        }
-
-      i += 3;  /* Advance past '://' */
-
-      path_ancestor_len = get_path_ancestor_length(path1 + i, path2 + i, 
-                                                   pool);
-
-      if (path_ancestor_len == 0)
-        return apr_pmemdup(pool, SVN_EMPTY_PATH, sizeof(SVN_EMPTY_PATH));
-      else
-        return apr_pstrndup(pool, path1, path_ancestor_len + i); 
-    }
-
-  else if ((! path1_is_url) && (! path2_is_url))
-    { 
-      return apr_pstrndup(pool, path1, 
-                          get_path_ancestor_length(path1, path2, pool));
-    }
-
-  else
-    {
-      /* A URL and a non-URL => no common prefix */
-      return apr_pmemdup(pool, SVN_EMPTY_PATH, sizeof(SVN_EMPTY_PATH));
-    }
+  return svn_uri_get_longest_ancestor(path1, path2, pool);
 }
 
-
 const char *
 svn_path_is_child(const char *path1,
                   const char *path2,
                   apr_pool_t *pool)
 {
-  apr_size_t i;
-
-  /* assert (is_canonical (path1, strlen (path1)));  ### Expensive strlen */
-  /* assert (is_canonical (path2, strlen (path2)));  ### Expensive strlen */
-
-  /* Allow "" and "foo" to be parent/child */
-  if (SVN_PATH_IS_EMPTY(path1))               /* "" is the parent  */
-    {
-      if (SVN_PATH_IS_EMPTY(path2)            /* "" not a child    */
-          || path2[0] == '/')                  /* "/foo" not a child */
-        return NULL;
-      else
-        return apr_pstrdup(pool, path2);      /* everything else is child */
-    }
-
-  /* Reach the end of at least one of the paths.  How should we handle
-     things like path1:"foo///bar" and path2:"foo/bar/baz"?  It doesn't
-     appear to arise in the current Subversion code, it's not clear to me
-     if they should be parent/child or not. */
-  for (i = 0; path1[i] && path2[i]; i++)
-    if (path1[i] != path2[i])
-      return NULL;
-
-  /* There are two cases that are parent/child
-          ...      path1[i] == '\0'
-          .../foo  path2[i] == '/'
-      or
-          /        path1[i] == '\0'
-          /foo     path2[i] != '/'
-  */
-  if (path1[i] == '\0' && path2[i])
-    {
-      if (path2[i] == '/')
-        return apr_pstrdup(pool, path2 + i + 1);
-      else if (i == 1 && path1[0] == '/')
-        return apr_pstrdup(pool, path2 + 1);
-    }
-
-  /* Otherwise, path2 isn't a child. */
-  return NULL;
+  return svn_uri_is_child(path1, path2, pool);
 }
 
+
 svn_boolean_t
 svn_path_is_ancestor(const char *path1, const char *path2)
 {
-  apr_size_t path1_len = strlen(path1);
-
-  /* If path1 is empty and path2 is not absoulte, then path1 is an ancestor. */
-  if (SVN_PATH_IS_EMPTY(path1))
-    return *path2 != '/';
-
-  /* If path1 is a prefix of path2, then:
-     - If path1 ends in a path separator,
-     - If the paths are of the same length
-     OR
-     - path2 starts a new path component after the common prefix,
-     then path1 is an ancestor. */
-  if (strncmp(path1, path2, path1_len) == 0)
-    return path1[path1_len - 1] == '/'
-      || (path2[path1_len] == '/' || path2[path1_len] == '\0');
-
-  return FALSE;
+  return svn_uri_is_ancestor(path1, path2);
 }
+
+
 apr_array_header_t *
 svn_path_decompose(const char *path,
                    apr_pool_t *pool)
 {
   apr_size_t i, oldi;
 
-  apr_array_header_t *components = 
+  apr_array_header_t *components =
     apr_array_make(pool, 1, sizeof(const char *));
 
-  /* assert (is_canonical (path, strlen (path)));  ### Expensive strlen */
+  assert(svn_path_is_canonical(path, pool));
 
   if (SVN_PATH_IS_EMPTY(path))
     return components;  /* ### Should we return a "" component? */
@@ -651,7 +501,7 @@ svn_path_decompose(const char *path,
     {
       char dirsep = '/';
 
-      *((const char **) apr_array_push(components))
+      APR_ARRAY_PUSH(components, const char *)
         = apr_pstrmemdup(pool, &dirsep, sizeof(dirsep));
 
       i++;
@@ -665,9 +515,9 @@ svn_path_decompose(const char *path,
       if ((path[i] == '/') || (path[i] == '\0'))
         {
           if (SVN_PATH_IS_PLATFORM_EMPTY(path + oldi, i - oldi))
-            *((const char **) apr_array_push(components)) = SVN_EMPTY_PATH;
+            APR_ARRAY_PUSH(components, const char *) = SVN_EMPTY_PATH;
           else
-            *((const char **) apr_array_push(components))
+            APR_ARRAY_PUSH(components, const char *)
               = apr_pstrmemdup(pool, path + oldi, i - oldi);
 
           i++;
@@ -682,10 +532,57 @@ svn_path_decompose(const char *path,
 }
 
 
+const char *
+svn_path_compose(const apr_array_header_t *components,
+                 apr_pool_t *pool)
+{
+  apr_size_t *lengths = apr_palloc(pool, components->nelts*sizeof(*lengths));
+  apr_size_t max_length = components->nelts;
+  char *path;
+  char *p;
+  int i;
+
+  /* Get the length of each component so a total length can be
+     calculated. */
+  for (i = 0; i < components->nelts; ++i)
+    {
+      apr_size_t l = strlen(APR_ARRAY_IDX(components, i, const char *));
+      lengths[i] = l;
+      max_length += l;
+    }
+
+  path = apr_palloc(pool, max_length + 1);
+  p = path;
+
+  for (i = 0; i < components->nelts; ++i)
+    {
+      /* Append a '/' to the path.  Handle the case with an absolute
+         path where a '/' appears in the first component.  Only append
+         a '/' if the component is the second component that does not
+         follow a "/" first component; or it is the third or later
+         component. */
+      if (i > 1 ||
+          (i == 1 && strcmp("/", APR_ARRAY_IDX(components,
+                                               0,
+                                               const char *)) != 0))
+        {
+          *p++ = '/';
+        }
+
+      memcpy(p, APR_ARRAY_IDX(components, i, const char *), lengths[i]);
+      p += lengths[i];
+    }
+
+  *p = '\0';
+
+  return path;
+}
+
+
 svn_boolean_t
 svn_path_is_single_path_component(const char *name)
 {
-  /* assert (is_canonical (name, strlen (name)));  ### Expensive strlen */
+  assert(is_canonical(name, strlen(name)));
 
   /* Can't be empty or `..'  */
   if (SVN_PATH_IS_EMPTY(name)
@@ -702,24 +599,55 @@ svn_path_is_single_path_component(const char *name)
 
 
 svn_boolean_t
-svn_path_is_backpath_present(const char *path)
+svn_path_is_dotpath_present(const char *path)
 {
-  int len = strlen(path);
-  
-  if (! strcmp(path, ".."))
+  int len;
+
+  /* The empty string does not have a dotpath */
+  if (path[0] == '\0')
+    return FALSE;
+
+  /* Handle "." or a leading "./" */
+  if (path[0] == '.' && (path[1] == '\0' || path[1] == '/'))
     return TRUE;
 
-  if (! strncmp(path, "../", 3))
+  /* Paths of length 1 (at this point) have no dotpath present. */
+  if (path[1] == '\0')
+    return FALSE;
+
+  /* If any segment is "/./", then a dotpath is present. */
+  if (strstr(path, "/./") != NULL)
     return TRUE;
-  
+
+  /* Does the path end in "/." ? */
+  len = strlen(path);
+  return path[len - 2] == '/' && path[len - 1] == '.';
+}
+
+svn_boolean_t
+svn_path_is_backpath_present(const char *path)
+{
+  int len;
+
+  /* 0 and 1-length paths do not have a backpath */
+  if (path[0] == '\0' || path[1] == '\0')
+    return FALSE;
+
+  /* Handle ".." or a leading "../" */
+  if (path[0] == '.' && path[1] == '.' && (path[2] == '\0' || path[2] == '/'))
+    return TRUE;
+
+  /* Paths of length 2 (at this point) have no backpath present. */
+  if (path[2] == '\0')
+    return FALSE;
+
+  /* If any segment is "..", then a backpath is present. */
   if (strstr(path, "/../") != NULL)
     return TRUE;
 
-  if (len >= 3
-      && (! strncmp(path + len - 3, "/..", 3)))
-    return TRUE;
-
-  return FALSE;
+  /* Does the path end in "/.." ? */
+  len = strlen(path);
+  return path[len - 3] == '/' && path[len - 2] == '.' && path[len - 1] == '.';
 }
 
 
@@ -734,9 +662,10 @@ skip_uri_scheme(const char *path)
 {
   apr_size_t j;
 
-  for (j = 0; path[j]; ++j)
-    if (path[j] == ':' || path[j] == '/')
-       break;
+  /* A scheme is terminated by a : and cannot contain any /'s. */
+  for (j = 0; path[j] && path[j] != ':'; ++j)
+    if (path[j] == '/')
+      return NULL;
 
   if (j > 0 && path[j] == ':' && path[j+1] == '/' && path[j+2] == '/')
     return path + j + 3;
@@ -744,19 +673,20 @@ skip_uri_scheme(const char *path)
   return NULL;
 }
 
-svn_boolean_t 
+
+svn_boolean_t
 svn_path_is_url(const char *path)
 {
   /* ### This function is reaaaaaaaaaaaaaally stupid right now.
      We're just going to look for:
- 
+
         (scheme)://(optional_stuff)
 
      Where (scheme) has no ':' or '/' characters.
 
      Someday it might be nice to have an actual URI parser here.
   */
-  return skip_uri_scheme(path) ? TRUE : FALSE;
+  return skip_uri_scheme(path) != NULL;
 }
 
 
@@ -764,7 +694,7 @@ svn_path_is_url(const char *path)
 /* Here is the BNF for path components in a URI. "pchar" is a
    character in a path component.
 
-      pchar       = unreserved | escaped | 
+      pchar       = unreserved | escaped |
                     ":" | "@" | "&" | "=" | "+" | "$" | ","
       unreserved  = alphanum | mark
       mark        = "-" | "_" | "." | "!" | "~" | "*" | "'" | "(" | ")"
@@ -772,7 +702,7 @@ svn_path_is_url(const char *path)
    Note that "escaped" doesn't really apply to what users can put in
    their paths, so that really means the set of characters is:
 
-      alphanum | mark | ":" | "@" | "&" | "=" | "+" | "$" | "," 
+      alphanum | mark | ":" | "@" | "&" | "=" | "+" | "$" | ","
 */
 static const char uri_char_validity[256] = {
   0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,
@@ -799,7 +729,8 @@ static const char uri_char_validity[256] = {
   0, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0, 0, 0, 0, 0,
 };
 
-svn_boolean_t 
+
+svn_boolean_t
 svn_path_is_uri_safe(const char *path)
 {
   apr_size_t i;
@@ -836,11 +767,12 @@ svn_path_is_uri_safe(const char *path)
         {
           return FALSE;
         }
-    } 
+    }
 
   return TRUE;
 }
-  
+
+
 /* URI-encode each character c in PATH for which TABLE[c] is 0.
    If no encoding was needed, return PATH, else return a new string allocated
    in POOL. */
@@ -851,7 +783,7 @@ uri_escape(const char *path, const char table[], apr_pool_t *pool)
   apr_size_t i, copied = 0;
   int c;
 
-  retstr = svn_stringbuf_create("", pool);
+  retstr = svn_stringbuf_create_ensure(strlen(path), pool);
   for (i = 0; path[i]; i++)
     {
       c = (unsigned char)path[i];
@@ -865,9 +797,9 @@ uri_escape(const char *path, const char table[], apr_pool_t *pool)
       /* First things first, copy all the good stuff that we haven't
          yet copied into our output buffer. */
       if (i - copied)
-        svn_stringbuf_appendbytes(retstr, path + copied, 
+        svn_stringbuf_appendbytes(retstr, path + copied,
                                   i - copied);
-      
+
       /* Now, sprintf() in our escaped character, making sure our
          buffer is big enough to hold the '%' and two digits.  We cast
          the C to unsigned char here because the 'X' format character
@@ -896,6 +828,7 @@ uri_escape(const char *path, const char table[], apr_pool_t *pool)
   return retstr->data;
 }
 
+
 const char *
 svn_path_uri_encode(const char *path, apr_pool_t *pool)
 {
@@ -937,7 +870,7 @@ svn_path_uri_from_iri(const char *iri, apr_pool_t *pool)
   return uri_escape(iri, iri_escape_chars, pool);
 }
 
-const char uri_autoescape_chars[256] = {
+static const char uri_autoescape_chars[256] = {
   1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1,
   1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1,
   0, 1, 0, 1, 1, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1,
@@ -975,10 +908,8 @@ svn_path_uri_decode(const char *path, apr_pool_t *pool)
   apr_size_t i;
   svn_boolean_t query_start = FALSE;
 
-  retstr = svn_stringbuf_create("", pool);
-
   /* avoid repeated realloc */
-  svn_stringbuf_ensure(retstr, strlen(path) + 1); 
+  retstr = svn_stringbuf_create_ensure(strlen(path) + 1, pool);
 
   retstr->len = 0;
   for (i = 0; path[i]; i++)
@@ -1017,12 +948,11 @@ svn_path_uri_decode(const char *path, apr_pool_t *pool)
 
 
 const char *
-svn_path_url_add_component(const char *url,
-                           const char *component,
-                           apr_pool_t *pool)
+svn_path_url_add_component2(const char *url,
+                            const char *component,
+                            apr_pool_t *pool)
 {
-  /* URL can have trailing '/' */
-  url = svn_path_canonicalize(url, pool);
+  assert(svn_path_is_canonical(url, pool));
 
   return svn_path_join(url, svn_path_uri_encode(component, pool), pool);
 }
@@ -1032,34 +962,13 @@ svn_path_get_absolute(const char **pabsolute,
                       const char *relative,
                       apr_pool_t *pool)
 {
-  char *buffer;
-  apr_status_t apr_err;
-  const char *path_apr;
-
-  SVN_ERR(svn_path_cstring_from_utf8
-          (&path_apr, svn_path_canonicalize(relative, pool), pool));
-
-  if (svn_path_is_url(path_apr))
+  if (svn_path_is_url(relative))
     {
-      buffer = apr_pstrdup(pool, path_apr);
-    }
-  else
-    {
-      apr_err = apr_filepath_merge(&buffer, NULL,
-                                   path_apr,
-                                   (APR_FILEPATH_NOTRELATIVE
-                                    | APR_FILEPATH_TRUENAME),
-                                   pool);
-
-      if (apr_err)
-        return svn_error_createf(SVN_ERR_BAD_FILENAME, NULL,
-                                 _("Couldn't determine absolute path of '%s'"), 
-                                 svn_path_local_style(relative, pool));
+      *pabsolute = apr_pstrdup(pool, relative);
+      return SVN_NO_ERROR;
     }
 
-  SVN_ERR(svn_path_cstring_to_utf8(pabsolute, buffer, pool));
-  *pabsolute = svn_path_canonicalize(*pabsolute, pool);
-  return SVN_NO_ERROR;
+  return svn_dirent_get_absolute(pabsolute, relative, pool);
 }
 
 
@@ -1072,7 +981,7 @@ svn_path_split_if_file(const char *path,
   apr_finfo_t finfo;
   svn_error_t *err;
 
-  /* assert (is_canonical (path, strlen (path)));  ### Expensive strlen */
+  SVN_ERR_ASSERT(svn_path_is_canonical(path, pool));
 
   err = svn_io_stat(&finfo, path, APR_FINFO_TYPE, pool);
   if (err && ! APR_STATUS_IS_ENOENT(err->apr_err))
@@ -1080,8 +989,7 @@ svn_path_split_if_file(const char *path,
 
   if (err || finfo.filetype == APR_REG)
     {
-      if (err)
-        svn_error_clear(err);
+      svn_error_clear(err);
       svn_path_split(path, pdirectory, pfile, pool);
     }
   else if (finfo.filetype == APR_DIR)
@@ -1089,7 +997,7 @@ svn_path_split_if_file(const char *path,
       *pdirectory = path;
       *pfile = SVN_EMPTY_PATH;
     }
-  else 
+  else
     {
       return svn_error_createf(SVN_ERR_BAD_FILENAME, NULL,
                                _("'%s' is neither a file nor a directory name"),
@@ -1103,89 +1011,23 @@ svn_path_split_if_file(const char *path,
 const char *
 svn_path_canonicalize(const char *path, apr_pool_t *pool)
 {
-  char *canon, *dst;
-  const char *src;
-  apr_size_t seglen;
-  apr_size_t canon_segments = 0;
-  svn_boolean_t uri;
-
-  dst = canon = apr_pcalloc(pool, strlen(path) + 1);
-
-  /* Copy over the URI scheme if present. */
-  src = skip_uri_scheme(path);
-  if (src)
-    {
-      uri = TRUE;
-      memcpy(dst, path, src - path);
-      dst += (src - path);
-    }
-  else
-    {
-      uri = FALSE;
-      src = path;
-    }
-
-  /* If this is an absolute path, then just copy over the initial
-     separator character. */
-  if (*src == '/')
-    {
-      *(dst++) = *(src++);
-
 #if defined(WIN32) || defined(__CYGWIN__)
-      /* On Windows permit two leading separator characters which means an
-       * UNC path.  However, a double slash in a URI after the scheme is never
-       * valid. */
-      if (!uri && *src == '/')
-        *(dst++) = *(src++);
-#endif /* WIN32 or Cygwin */
-      
-    }
-
-  while (*src)
-    {
-      /* Parse each segment, find the closing '/' */
-      const char *next = src;
-      while (*next && (*next != '/'))
-        ++next;
-
-      seglen = next - src;
-
-      if (seglen == 0 || (seglen == 1 && src[0] == '.'))
-        {
-          /* Noop segment, so do nothing. */
-        }
-      else
-        {
-          /* An actual segment, append it to the destination path */
-          if (*next)
-            seglen++;
-          memcpy(dst, src, seglen);
-          dst += seglen;
-          canon_segments++;
-        }
-
-      /* Skip over trailing slash to the next segment. */
-      src = next;
-      if (*src)
-        src++;
-    }
-
-  /* Remove the trailing slash. */
-  if ((canon_segments > 0 || uri) && *(dst - 1) == '/')
-    dst--;
-  
-  *dst = '\0';
-
-#if defined(WIN32) || defined(__CYGWIN__)
-  /* Skip leading double slashes when there are less than 2
-   * canon segments. UNC paths *MUST* have two segments. */
-  if (canon_segments < 2 && canon[0] == '/' && canon[1] == '/')
-    return canon + 1;
-#endif /* WIN32 or Cygwin */
-
-  return canon;
+  if (path[0] == '/' && path[1] == '/')
+    return svn_dirent_canonicalize(path, pool);
+#endif
+  return svn_uri_canonicalize(path, pool);
 }
 
+svn_boolean_t
+svn_path_is_canonical(const char *path, apr_pool_t *pool)
+{
+  return svn_uri_is_canonical(path, pool)
+#if defined(WIN32) || defined(__CYGWIN__)
+         || (path[0] == '/' && path[1] == '/' &&
+             svn_dirent_is_canonical(path, pool))
+#endif
+         ;
+}
 
 
 /** Get APR's internal path encoding. */
@@ -1241,6 +1083,62 @@ svn_path_cstring_to_utf8(const char **path_utf8,
     return svn_utf_cstring_to_utf8(path_utf8, path_apr, pool);
 }
 
+
+/* Return a copy of PATH, allocated from POOL, for which control
+   characters have been escaped using the form \NNN (where NNN is the
+   octal representation of the byte's ordinal value).  */
+static const char *
+illegal_path_escape(const char *path, apr_pool_t *pool)
+{
+  svn_stringbuf_t *retstr;
+  apr_size_t i, copied = 0;
+  int c;
+
+  /* At least one control character:
+      strlen - 1 (control) + \ + N + N + N + null . */
+  retstr = svn_stringbuf_create_ensure(strlen(path) + 4, pool);
+  for (i = 0; path[i]; i++)
+    {
+      c = (unsigned char)path[i];
+      if (! svn_ctype_iscntrl(c))
+        continue;
+
+      /* If we got here, we're looking at a character that isn't
+         supported by the (or at least, our) URI encoding scheme.  We
+         need to escape this character.  */
+
+      /* First things first, copy all the good stuff that we haven't
+         yet copied into our output buffer. */
+      if (i - copied)
+        svn_stringbuf_appendbytes(retstr, path + copied,
+                                  i - copied);
+
+      /* Make sure buffer is big enough for '\' 'N' 'N' 'N' null */
+      svn_stringbuf_ensure(retstr, retstr->len + 5);
+      /*### The backslash separator doesn't work too great with Windows,
+         but it's what we'll use for consistency with invalid utf8
+         formatting (until someone has a better idea) */
+      sprintf(retstr->data + retstr->len, "\\%03o", (unsigned char)c);
+      retstr->len += 4;
+
+      /* Finally, update our copy counter. */
+      copied = i + 1;
+    }
+
+  /* If we didn't encode anything, we don't need to duplicate the string. */
+  if (retstr->len == 0)
+    return path;
+
+  /* Anything left to copy? */
+  if (i - copied)
+    svn_stringbuf_appendbytes(retstr, path + copied, i - copied);
+
+  /* retstr is null-terminated either by sprintf or the svn_stringbuf
+     functions. */
+
+  return retstr->data;
+}
+
 svn_error_t *
 svn_path_check_valid(const char *path, apr_pool_t *pool)
 {
@@ -1253,10 +1151,53 @@ svn_path_check_valid(const char *path, apr_pool_t *pool)
           return svn_error_createf
             (SVN_ERR_FS_PATH_SYNTAX, NULL,
              _("Invalid control character '0x%02x' in path '%s'"),
-             *c,
-             svn_path_local_style(path, pool));
+             (unsigned char)*c,
+             illegal_path_escape(svn_path_local_style(path, pool), pool));
         }
     }
 
   return SVN_NO_ERROR;
 }
+
+void
+svn_path_splitext(const char **path_root,
+                  const char **path_ext,
+                  const char *path,
+                  apr_pool_t *pool)
+{
+  const char *last_dot, *last_slash;
+
+  /* Easy out -- why do all the work when there's no way to report it? */
+  if (! (path_root || path_ext))
+    return;
+
+  /* Do we even have a period in this thing?  And if so, is there
+     anything after it?  We look for the "rightmost" period in the
+     string. */
+  last_dot = strrchr(path, '.');
+  if (last_dot && (last_dot + 1 != '\0'))
+    {
+      /* If we have a period, we need to make sure it occurs in the
+         final path component -- that there's no path separator
+         between the last period and the end of the PATH -- otherwise,
+         it doesn't count.  Also, we want to make sure that our period
+         isn't the first character of the last component. */
+      last_slash = strrchr(path, '/');
+      if ((last_slash && (last_dot > (last_slash + 1)))
+          || ((! last_slash) && (last_dot > path)))
+        {
+          if (path_root)
+            *path_root = apr_pstrmemdup(pool, path,
+                                        (last_dot - path + 1) * sizeof(*path));
+          if (path_ext)
+            *path_ext = apr_pstrdup(pool, last_dot + 1);
+          return;
+        }
+    }
+  /* If we get here, we never found a suitable separator character, so
+     there's no split. */
+  if (path_root)
+    *path_root = apr_pstrdup(pool, path);
+  if (path_ext)
+    *path_ext = "";
+}
